@@ -1,4 +1,4 @@
-# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
+# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
 # All rights reserved.
 #
 # This software is provided without warranty under the terms of the BSD
@@ -27,7 +27,6 @@ from traits.api import (
     Any,
     Event,
     HasStrictTraits,
-    Instance,
     Int,
     List,
     Str,
@@ -156,9 +155,9 @@ class _TraitsChangeCollector(HasStrictTraits):
     # Private list of events.
     events = List(Any)
 
-    # Lock used to allow access to events by multiple threads
+    # Private lock used to allow access to events by multiple threads
     # simultaneously.
-    _lock = Instance(threading.Lock, ())
+    _lock = Any()
 
     def __init__(self, **traits):
         if "trait" in traits:
@@ -168,7 +167,13 @@ class _TraitsChangeCollector(HasStrictTraits):
             )
             warnings.warn(message, DeprecationWarning, stacklevel=2)
             traits["trait_name"] = value
-        super(_TraitsChangeCollector, self).__init__(**traits)
+        super().__init__(**traits)
+
+        # We assign the lock eagerly rather than depending on a lazy default,
+        # since we want to be sure that the lock is created (a) only once,
+        # and (b) on the main thread. Similarly for the events list.
+        self._lock = threading.Lock()
+        self.events = []
 
     def start_collecting(self):
         self.obj.on_trait_change(self._event_handler, self.trait_name)
@@ -454,6 +459,29 @@ class UnittestTools(object):
 
     @contextlib.contextmanager
     def _catch_warnings(self):
+        """
+        Replacement for warnings.catch_warnings.
+
+        This method wraps warnings.catch_warnings, takes care to
+        reset the warning registry before entering the with context,
+        and ensures that DeprecationWarnings are always emitted.
+
+        The hack to reset the warning registry is no longer needed in
+        Python 3.4 and later. See http://bugs.python.org/issue4180 for
+        more background.
+
+        .. deprecated:: 6.2
+            Use :func:`warnings.catch_warnings` instead.
+
+        """
+        warnings.warn(
+            (
+                "The _catch_warnings method is deprecated. "
+                "Use warnings.catch_warnings instead."
+            ),
+            DeprecationWarning,
+        )
+
         # Ugly hack copied from the core Python code (see
         # Lib/test/test_support.py) to reset the warnings registry
         # for the module making use of this context manager.
@@ -475,8 +503,10 @@ class UnittestTools(object):
         for testing uses of traits.util.deprecated.deprecated.
 
         """
-        with self._catch_warnings() as w:
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always", DeprecationWarning)
             yield w
+
         self.assertGreater(
             len(w),
             0,
@@ -490,8 +520,10 @@ class UnittestTools(object):
         for testing uses of traits.util.deprecated.deprecated.
 
         """
-        with self._catch_warnings() as w:
+        with warnings.catch_warnings(record=True) as w:
+            warnings.simplefilter("always", DeprecationWarning)
             yield w
+
         self.assertEqual(
             len(w),
             0,
