@@ -1,4 +1,4 @@
-# (C) Copyright 2005-2020 Enthought, Inc., Austin, TX
+# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
 # All rights reserved.
 #
 # This software is provided without warranty under the terms of the BSD
@@ -22,7 +22,18 @@ from traits.has_traits import (
 from traits.testing.optional_dependencies import numpy, requires_numpy
 from traits.trait_errors import TraitError
 from traits.trait_type import TraitType
-from traits.trait_types import Bool, DelegatesTo, Either, Instance, Int, List
+from traits.trait_types import (
+    Bool,
+    DelegatesTo,
+    Dict,
+    Either,
+    Instance,
+    Int,
+    List,
+    Set,
+    Str,
+    Union,
+)
 
 if numpy is not None:
     from traits.trait_numeric import Array
@@ -161,7 +172,11 @@ class TestRegression(unittest.TestCase):
         _create_subclass()
         _create_subclass()
         gc.collect()
-        self.assertEqual(previous_subclasses, HasTraits.__subclasses__())
+
+        # In Python < 3.6, we can end up seeing the same subclasses but in
+        # a different order, so use assertCountEqual rather than assertEqual.
+        # Ref: enthought/traits#1282.
+        self.assertCountEqual(previous_subclasses, HasTraits.__subclasses__())
 
     def test_leaked_property_tuple(self):
         """ the property ctrait constructor shouldn't leak a tuple. """
@@ -170,7 +185,7 @@ class TestRegression(unittest.TestCase):
             prop = Property()
 
         a = A()
-        self.assertEqual(sys.getrefcount(a.trait("prop").property()), 1)
+        self.assertEqual(sys.getrefcount(a.trait("prop").property_fields), 1)
 
     def test_delegate_initializer(self):
         mess = DelegateMess()
@@ -289,3 +304,146 @@ class TestRegression(unittest.TestCase):
 
         with self.assertRaises(ZeroDivisionError):
             a.bar = "foo"
+
+
+class NestedContainerClass(HasTraits):
+    # Used in regression test for changes to nested containers
+
+    # Nested list
+    list_of_list = List(List)
+
+    # enthought/traits#281
+    dict_of_list = Dict(Str, List(Str))
+
+    # Similar to enthought/traits#281
+    dict_of_union_none_or_list = Dict(Str, Union(List(), None))
+
+    # Nested dict
+    # enthought/traits#25
+    list_of_dict = List(Dict)
+
+    dict_of_dict = Dict(Str, Dict)
+
+    dict_of_union_none_or_dict = Dict(Str, Union(Dict(), None))
+
+    # Nested set
+    list_of_set = List(Set)
+
+    dict_of_set = Dict(Str, Set)
+
+    dict_of_union_none_or_set = Dict(Str, Union(Set(), None))
+
+
+class TestRegressionNestedContainerEvent(unittest.TestCase):
+    """ Regression tests for enthought/traits#281 and enthought/traits#25
+    """
+
+    def setUp(self):
+        self.events = []
+
+        def change_handler(*args):
+            self.events.append(args)
+
+        self.change_handler = change_handler
+
+    def test_modify_list_in_dict(self):
+        # Regression test for enthought/traits#281
+        instance = NestedContainerClass(dict_of_list={"name": []})
+
+        try:
+            instance.dict_of_list["name"].append("word")
+        except Exception:
+            self.fail("Mutating a nested list should not fail.")
+
+    def test_modify_list_in_dict_wrapped_in_union(self):
+        instance = NestedContainerClass(
+            dict_of_union_none_or_list={"name": []},
+        )
+        try:
+            instance.dict_of_union_none_or_list["name"].append("word")
+        except Exception:
+            self.fail("Mutating a nested list should not fail.")
+
+    def test_modify_list_in_list_no_events(self):
+        instance = NestedContainerClass(list_of_list=[[]])
+        instance.on_trait_change(self.change_handler, "list_of_list_items")
+
+        instance.list_of_list[0].append(1)
+        self.assertEqual(len(self.events), 0, "Expected no events.")
+
+    def test_modify_dict_in_list(self):
+        instance = NestedContainerClass(list_of_dict=[{}])
+
+        try:
+            instance.list_of_dict[0]["key"] = 1
+        except Exception:
+            self.fail("Mutating a nested dict should not fail.")
+
+    def test_modify_dict_in_list_with_new_value(self):
+        instance = NestedContainerClass(list_of_dict=[{}])
+
+        instance.list_of_dict.append(dict())
+        try:
+            instance.list_of_dict[-1]["key"] = 1
+        except Exception:
+            self.fail("Mutating a nested dict should not fail.")
+
+    def test_modify_dict_in_dict_no_events(self):
+        # Related to enthought/traits#25
+        instance = NestedContainerClass(dict_of_dict={"1": {"2": 2}})
+        instance.on_trait_change(self.change_handler, "dict_of_dict_items")
+
+        instance.dict_of_dict["1"]["3"] = 3
+
+        self.assertEqual(len(self.events), 0, "Expected no events.")
+
+    def test_modify_dict_in_union_in_dict_no_events(self):
+        instance = NestedContainerClass(
+            dict_of_union_none_or_dict={"1": {"2": 2}},
+        )
+        instance.on_trait_change(
+            self.change_handler, "dict_of_union_none_or_dict_items")
+
+        instance.dict_of_union_none_or_dict["1"]["3"] = 3
+
+        self.assertEqual(len(self.events), 0, "Expected no events.")
+
+    def test_modify_set_in_list(self):
+        instance = NestedContainerClass(list_of_set=[set()])
+        try:
+            instance.list_of_set[0].add(1)
+        except Exception:
+            self.fail("Mutating a nested set should not fail.")
+
+    def test_modify_set_in_list_with_new_value(self):
+        instance = NestedContainerClass(list_of_set=[])
+        instance.list_of_set.append(set())
+        try:
+            instance.list_of_set[0].add(1)
+        except Exception:
+            self.fail("Mutating a nested set should not fail.")
+
+    def test_modify_set_in_dict(self):
+        instance = NestedContainerClass(dict_of_set={"1": set()})
+        try:
+            instance.dict_of_set["1"].add(1)
+        except Exception:
+            self.fail("Mutating a nested set should not fail.")
+
+    def test_modify_set_in_union_in_dict(self):
+        instance = NestedContainerClass(
+            dict_of_union_none_or_set={"1": set()}
+        )
+        try:
+            instance.dict_of_union_none_or_set["1"].add(1)
+        except Exception:
+            self.fail("Mutating a nested set should not fail.")
+
+    def test_modify_nested_set_no_events(self):
+        instance = NestedContainerClass(list_of_set=[set()])
+        instance.on_trait_change(
+            self.change_handler, "list_of_set_items")
+
+        instance.list_of_set[0].add(1)
+
+        self.assertEqual(len(self.events), 0, "Expected no events.")
